@@ -11,8 +11,8 @@
  * Library General Public License for more details.
  *
  * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Cambridge, MA 02139, USA.
  */
 #ifndef __GNOME_MESSAGE_BOX_H__
 #define __GNOME_MESSAGE_BOX_H__
@@ -48,15 +48,20 @@ struct _GnomeMessageBoxClass
 };
 
 
-guint      gnome_message_box_get_type       (void);
-GtkWidget* gnome_message_box_new (const gchar           *message,
-				  const gchar           *messagebox_type,
-				  ...);
+guint      gnome_message_box_get_type    (void);
+GtkWidget* gnome_message_box_new         (const gchar           *message,
+					  const gchar           *messagebox_type,
+					  ...);
 
-/* Deprecated in favor of gnome_dialog_ variants. Don't use. */
-void       gnome_message_box_set_modal      (GnomeMessageBox *messagebox);
-void       gnome_message_box_set_default    (GnomeMessageBox *messagebox,
-					     gint            button);
+GtkWidget* gnome_message_box_newv        (const gchar           *message,
+					  const gchar           *messagebox_type,
+					  const gchar 	      **buttons);
+/* Deprecated in favor of gtk_window's version. Don't use. */
+void       gnome_message_box_set_modal   (GnomeMessageBox *messagebox);
+
+/* Deprecated in favor of gnome_dialog_ variant. Don't use. */
+void       gnome_message_box_set_default (GnomeMessageBox *messagebox,
+					  gint            button);
 
 END_GNOME_DECLS
 
